@@ -3,6 +3,7 @@ from graphviz import Digraph
 from IPython.display import display
 from networkx.readwrite import json_graph
 from urllib.parse import quote
+import fnmatch
 import json
 import networkx as nx
 
@@ -40,6 +41,17 @@ class ER:
         self.nodes = dict()
         self.nodesInfoDict = dict()
         self.relations = list()
+
+        self.__default_scores = {
+            'missing_object': 1,
+            'missing_property': {
+                str(NodeType.NODE): 0.5,
+                str(NodeType.ATTRIBUTE): 0.25,
+                str(NodeType.RELATION): 0.5,
+                str(NodeType.COMPOSED_ATTRIBUTE): 0.125,
+                str(NodeType.IS_A): 0.25
+            }
+        }
         
     @classmethod
     def copyfrom(cls, diagram, engine='dot', edge_len=1.5, debug=False, graph_attr={}):
@@ -447,16 +459,7 @@ class ER:
         return self.graph
 
     def get_default_scores(self):
-        return {
-            'missing_node': 1,
-            'missing_property': {
-                str(NodeType.NODE): 0.5,
-                str(NodeType.ATTRIBUTE): 0.25,
-                str(NodeType.RELATION): 0.5,
-                str(NodeType.IS_A): 0.25
-            }
-        }
-        
+        return self.__default_scores        
 
     def _add_obj_node(self, obj):
         self.add_node(obj['label'], isMultiple = obj['isMultiple'], isWeak = obj['isWeak'])
@@ -545,37 +548,128 @@ class ER:
             if debugging:
                 print("  no scores provided, fallback to default.")
             scores = self.get_default_scores()
+
+
         for n1 in self.get_obj(label, node_type = node_type):
+            # ignore attributes, will be checked as part of NodeType.NODE
+            if n1["nodeType"] == str(NodeType.ATTRIBUTE) or n1["nodeType"] == str(NodeType.COMPOSED_ATTRIBUTE):
+                continue
+
             if debugging:
                 print(f" » testing {n1['label']}:")
-            if not otherGraph.has_node(n1["label"]):
+
+            distancePerProperty = scores['missing_property'][n1['nodeType']]
+
+            # check (by label) if object exists in other graph
+            if not otherGraph.has_obj(n1["label"], n1["nodeType"]):
+                # not found.
                 if debugging:
-                    print(f"   ✗ {scores['missing_object']:+.2f}, object '{n1['label']}' doesn't exist in other graph")
+                    print(f"   ✗ {scores['missing_object']:+.2f}, {n1['nodeType']} '{n1['label']}' doesn't exist in other graph")
                 dist += scores['missing_object']
-            else:
-                # node exists, check equality
-                n2 = otherGraph.get_obj(n1["label"], node_type)
-                if n1 == n2:
-                    if debugging:
-                        print(f"   ✓  {0:.2f}, match@{n1['label']}")
-                        print(f"   =  {dist:.2f}")
-                    continue
-                distancePerProperty = scores['missing_property'][n1['nodeType']]
 
-                for k, v in n1.items():
-                    if v != n2[k]:
+                # additionally substract points for missing node attributes
+                if n1["nodeType"] == str(NodeType.NODE):
+                    # get all attributes
+                    x = self.get_attr_and_comp(f"{n1['label']}.*")
+                    for y in x:
+                        distancePerProperty = scores['missing_property'][str(NodeType.ATTRIBUTE)]
                         if debugging:
-                            print(f"   ✗ {distancePerProperty:+.2f}, mismatch@{k}: {v} != {n2[k]} ")
+                            print(f"   ✗ {distancePerProperty:+.2f}, missing {NodeType.ATTRIBUTE} '{y['label']}' ")
                         dist += distancePerProperty
+            else:
+                # node exists, check equality and compare
+                if debugging: print(f"   ✓        exists")
+                n2 = otherGraph.get_obj(n1["label"], node_type)
 
-            if debugging:
-                print(f"   =  {dist:.2f}")
+                localDist = self.__compare_nodes(n1, n2, 
+                    scores, otherGraph, debugging)
+
+                dist += localDist
+
+                if debugging:
+                    print(f"   =  {dist:.2f}")
 
         if debugging:
             print(f" ---------------")
             print(f"   ∑  {dist:.2f}")        
         return dist
 
+    def __compare_properties(self, thisNode, otherNode, key, debugging):
+        #if debugging: 
+        #    print(f"comparing property[{thisNode['label']}.{key}]: {thisNode[key]} vs {otherNode[key]}")
+        thisValue = thisNode[key]
+        otherValue = otherNode[key]
+        
+        # RELATION: don't check "relation" and "relationLabel" strings
+        if thisNode['nodeType'] == str(NodeType.RELATION): 
+            if key == "relation":
+                return True
+            if key == "relationLabel":
+                return True
+        
+        # RELATION: don't differentiate between letters
+        if  (key == "fromEdgeLabel" or key == "toEdgeLabel") and (otherValue.isalpha() and thisValue.isalpha()):
+            #if debugging: 
+            #    print(f"   // ignoring literal mismatch in edge label {thisValue} vs {otherValue}")
+            return True
+        if thisValue != otherValue:
+            if debugging: 
+                print(f"property compare {key} fail {thisValue} vs {otherValue}")
+            return False
+        return True
+
+    def __compare_nodes(self, thisNode, otherNode, 
+                        scores, otherGraph, debugging):
+        #if thisNode['label'] != otherNode['label']:
+        #    if self.debug: print(f"node compare label fail {thisNode['label']} vs {otherNode['label']}")
+        #    return False
+
+        # TODO: refactor these items into separate classes with helper functions
+
+        localDist = 0
+
+        if thisNode["nodeType"] == str(NodeType.NODE):
+            # SPECIAL CASE: NodeType.NODE - compare node attributes
+            nodeAttributes = self.get_attr(f"{thisNode['label']}.*")
+            for localAttr in nodeAttributes:
+                distancePerProperty = scores['missing_property'][str(NodeType.ATTRIBUTE)]
+
+                # is attribute missing?
+                if not otherGraph.has_attr(localAttr['label']):
+                    if debugging:
+                        print(f"   ✗ {distancePerProperty:+.2f}, missing {NodeType.ATTRIBUTE} '{localAttr['label']}' ")
+                    localDist += distancePerProperty
+                else:
+                    # exists but check params (isWeak etc.)
+                    otherAttr = otherGraph.get_attr(localAttr['label'])
+                    for key, value in localAttr.items():
+                        if value != otherAttr[key]:
+                            if debugging:
+                                print(f"   ✗ {distancePerProperty:+.2f}, mismatch@{key}: {value} != {otherAttr[key]} ")
+                            localDist += distancePerProperty
+
+        # compare node properties (isWeak etc.)
+
+        # comparison property keys:
+        if thisNode['nodeType'] == str(NodeType.NODE):
+            propertyKeys = ['isMultiple', 'isWeak']
+        if thisNode['nodeType'] == str(NodeType.ATTRIBUTE) or thisNode['nodeType'] == str(NodeType.COMPOSED_ATTRIBUTE):
+            propertyKeys = ['attrLabel', 'parentLabel', 'isPK', 'isMultiple', 'isWeak', 'composedOf']
+        if thisNode['nodeType'] == str(NodeType.IS_A):
+            propertyKeys = ['relation', 'superClassLabel', 'superLabel', 'subLabel', 'isDisjunct', 'subClasses']
+        if thisNode['nodeType'] == str(NodeType.RELATION):
+            propertyKeys = ['relation', 'relationLabel', 'relationFrom', 'relationTo', 'fromEdgeLabel', 'toEdgeLabel', 'isWeak']
+        distancePerProperty = scores['missing_property'][thisNode['nodeType']]
+        for k in propertyKeys:
+            if not self.__compare_properties(thisNode, otherNode, k, debugging):
+                if debugging:
+                    print(f"   ✗ {distancePerProperty:+.2f}, property mismatch")
+                localDist += distancePerProperty
+            else:
+                continue
+        
+        return localDist
+
     def print_nodes(self):
         if self.debug:
             print("-> printing nodes")
@@ -588,7 +682,7 @@ class ER:
             print(e)
 
     def has_node(self, label):
-        return self.get_graph().has_node(label)
+        return self.has_obj(label, NodeType.NODE)#self.get_graph().has_node(label)
 
     def has_attr(self, label):
         return self.has_obj(label, NodeType.ATTRIBUTE)
@@ -596,6 +690,9 @@ class ER:
     def has_comp_attr(self, label):
         return self.has_obj(label, NodeType.COMPOSED_ATTRIBUTE)
 
+    def has_attr_or_comp(self, label):
+        return self.has_attr(label) or self.has_comp_attr(label)
+
     def has_rel(self, label):
         return self.has_obj(label, NodeType.RELATION)
 
@@ -616,12 +713,18 @@ class ER:
             If label is empty, gets all objects of this type.
             If node_type is NodeType.NOT_SPECIFIED, gets all objects in graph.
         """
+        if label.endswith(".*"):
+            label = label.partition("*")[0]
         obj_list = []
         for obj in list(self.get_graph().nodes(data=True)):
             if ( node_type == NodeType.NOT_SPECIFIED or obj[1]["nodeType"] == str(node_type) ):
-                if ( label == ""): 
+                if label == "": 
                     obj_list.append(obj[1]) # get all nodes of this type
-                if ( obj[1]["label"] == label ):
+                if label.endswith("."):
+                    #print("wildcard" + label)
+                    if obj[1]["label"].startswith(label):
+                        obj_list.append(obj[1])
+                if obj[1]["label"] == label:
                     return obj[1]
         return obj_list
 
@@ -645,6 +748,19 @@ class ER:
     def get_attr(self, label=""):
         return self.get_obj(label, NodeType.ATTRIBUTE)
 
+    def get_comp_attr(self, label=""):
+        return self.get_obj(label, NodeType.COMPOSED_ATTRIBUTE)
+
+    def get_attr_and_comp(self, label=""):
+        ret_list = []
+        attrs = self.get_obj(label, NodeType.ATTRIBUTE) 
+        comps = self.get_obj(label, NodeType.COMPOSED_ATTRIBUTE)
+        for i in attrs:
+            ret_list.append(i)
+        for i in comps:
+            ret_list.append(i)
+        return ret_list
+
     def get_isA(self, label=""):
         return self.get_obj(label, NodeType.IS_A)
 
